package tomp.xml.util;

import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.net.*;

public class XPP {
	
	static final String USAGE = "Usage: XPP [options] source_file_specification\t\n"
		+"\nwhere [options] may include:\n"
		
		+"[--verbose|-v] ... verbose mode\n"
		+"[--help|-h] ... prints this help\n"
		
		+"[--recurse|-r] directory ... recursively traverses directory, using pattern\n"
		+"[--pattern|-p] pattern ... \n\tspecifies pattern for directory traversal;\n\tif pattern not specified, all files are accepted\n"

		+"[--stdout|-s] ... sends output to stdout\n"
		+"[--outfile|-o] file ... saves output to file\n"
		+"[--destination|-d] directory ... saves output files into directory\n"
		+"[--always-rename|-a] ... \n\talways renames target files\n\tremoves the '.pp' extension\n\teven if saved into different directory\n"

		+"\nsource_file_specification may be in one of the following form:\n"
		+"filename(s) ... source specified by filename(s)\n"
		+"[-u|--url] url(s) ... source specified by URL(s)\n"
		+"[-l|--list] filename(s) ... file contains the list of sources\n"
		+"[-l|--list] [-u|--url] url(s) ... url(s) points to the list of sources\n"
		;

    Set sources = new HashSet();
        
    boolean alwaysRename = false;
    boolean verbose = false;
    boolean concat = false;
    boolean fromURL = false;
    boolean fromList = false;
    boolean stdout = false;
    
    boolean recurse = false;
    String recursePattern = null;
    
    File outFile = null;
    File destDir = null;
    
    Writer out = null;
    
    public static void main(String[] params)
       throws IOException
    {
		XPP xpp = new XPP();     
		
//		list(params);           

        int pi = 0;
        while (pi < params.length) {

            if (   params[pi].equalsIgnoreCase("--verbose") 
                || params[pi].equalsIgnoreCase("-v")) {
                	
                xpp.verbose = true;

            } else if (   params[pi].equalsIgnoreCase("--help")
                       || params[pi].equalsIgnoreCase("-h")) {
                       	
                usage(); 
                pi = params.length;
                break;

            } else if (   params[pi].equalsIgnoreCase("--url")
                       || params[pi].equalsIgnoreCase("-u")) {
                xpp.fromURL = true;

            } else if (   params[pi].equalsIgnoreCase("--stdout")
                       || params[pi].equalsIgnoreCase("-s")) {
                xpp.stdout = true;

            } else if (   params[pi].equalsIgnoreCase("--concat")
                       || params[pi].equalsIgnoreCase("-c")) {
                xpp.concat = true;
                
            } else if (   params[pi].equalsIgnoreCase("--recurse")
                       || params[pi].equalsIgnoreCase("-r")) {
                xpp.recurse = true;
                
            } else if (   params[pi].equalsIgnoreCase("--pattern")
                       || params[pi].equalsIgnoreCase("-p")) {
                pi++;
                xpp.recursePattern = params[pi];
                
            } else if (   params[pi].equalsIgnoreCase("--outfile")
                       || params[pi].equalsIgnoreCase("-o")) {
                pi++;
                xpp.outFile = new File(params[pi]);
                
            } else if (   params[pi].equalsIgnoreCase("--always-rename")
                       || params[pi].equalsIgnoreCase("-a")) {
                xpp.alwaysRename = true;
                
            } else if (   params[pi].equalsIgnoreCase("-d") 
                       || params[pi].equalsIgnoreCase("--dest")
                       || params[pi].equalsIgnoreCase("--destination") ) {
                pi++;
                xpp.destDir = new File(params[pi]);
                
                
            } else if (   params[pi].equalsIgnoreCase("-l") 
                       || params[pi].equalsIgnoreCase("--list")) {

				xpp.fromList = true;
				
            } else if (!params[pi].startsWith("-")) {

//				System.err.println("will add src="+params[pi]);
            	if (xpp.fromURL) {
					xpp.sources.add(new URL(params[pi]));
				} else {
					xpp.sources.add(new File(params[pi]));
				}
				
            } else {
	            throw new IllegalArgumentException(USAGE);
        	}
			
//			System.err.println("param["+pi+"] read.");
            pi++;
        }

        xpp.info();     
        xpp.process();        
    }

	public static void list(String[] p) {
		for (int i = 0; i < p.length; i++) {
			System.err.println("param "+i+".="+p[i]);
		}		
	}

	public static void usage() {
		System.err.println(USAGE);
	}

	protected void info()  throws IOException {
		msg("\nXPP executed with the following options...");
		msg("\n\talwaysRename="+alwaysRename);
		msg("\n\tconcat="+concat);
		msg("\n\tstdout="+stdout);
		msg("\n\tdestDir="+destDir);
		msg("\n\toutFile="+outFile);
		msg("\n\tfromURL="+fromURL);
		msg("\n\tfromList="+fromList);
		msg("\n\trecurse="+recurse);
		msg("\n\trecursePattern="+recursePattern);
		msg("\n\tverbose="+verbose);
		msg("\n\nStart preparing...");
		msg("\noriginal ");
		listSources();
		prepareSources();
//		msg("\nreal ");
//		listSources();
	}
	
	protected void listSources() {
		msg("sources"); // ("+sources.size()+"):");
		for (Iterator i = sources.iterator(); i.hasNext(); ) {
			Object o = i.next();
			msg("\n\t"+(o instanceof File ? "file ":"URL  ")+o);
		}
	}


	protected void prepareSources() throws IOException {
		
		if (fromList) {
			// set of list of sources -> convert to set of sources
			msg("\nsources specified by list ");
			
			Set ss = sources;
			sources = new HashSet();
			
			// get all sources
			for (Iterator i = ss.iterator(); i.hasNext();) {
				Object o = i.next();
			
				BufferedReader br = new BufferedReader(
						new InputStreamReader(createIS(o)));
						
//				System.err.println("BR="+br);						
				
				String line = br.readLine();
				try {
					while(line != null) {
						if (line.length() > 0) {
							File f = null;
							URL u = null;
							try { 
								u = new URL(line);
								sources.add(u);
							} catch(MalformedURLException rte) {
								f = new File(line);
								sources.add(f);
							}	
						}
						line = br.readLine();			
					}
				} catch (RuntimeException e) {
					e.printStackTrace();
				}
			}
		} else {
			msg("\nsources specified directly");
		}	

		if (recurse) {
			// sources now specify directories to be recursively traversed
			msg(" as directory & pattern");

			Set directories = sources;
			sources = new HashSet();
			FileFilter ff = new PatternFilter();
			
			for (Iterator i = directories.iterator(); i.hasNext(); ) {
				Object o = i.next();
				if (o instanceof File) { 
					File directory = (File)o;
					if (directory.isDirectory()) {
						traverse(sources, directory, ff);
					} else {
						msg("\n\tcannot traverse file (not directory) "+o);
					}
				} else {
					msg("\n\tcannot traverse URL "+o);
				}
			}
		} else {
			// sources now specify nonrecursive sources
			msg(" as files/URLs");

		}	
	
	}
	
	protected static FileFilter SF = new FileFilter() {
		public boolean accept (File f) {
			return f.isDirectory() && !f.getName().startsWith(".");
		}
	};	

	protected void traverse(Set s, File directory, FileFilter ff) {
		msg("\n\ttraversing directory '"+directory.getName());
		File[] list = directory.listFiles(ff);
		msg("', pattern=/"+recursePattern+"/, files matched="+list.length);
		/*
		for (int j = 0; j < list.length; j++) {
			String n = list[j].getName();
			System.out.println("\n"+n+" matches with "+ recursePattern+" ? "+n.matches(recursePattern));
		}
		*/

		for (int j = 0; j < list.length; j++) {
			s.add(list[j]);
		}
		File[] subdirs = directory.listFiles(SF);
		for (int j = 0; j < subdirs.length; j++) {
			traverse(s, subdirs[j], ff);
		}
	}
	
	protected class PatternFilter implements FileFilter {
		public boolean accept (File f) {
//			System.out.println("file f="+f+" testing filename with pattern "+recursePattern);
			if (f.isDirectory()) 
				return false;
			if (recursePattern == null) 
				return true;
			else 
				return f.getName().matches(recursePattern);
		}
	}
	
	protected File getOutfile(File in) {
		String name = in.getName();
		File parent = in.getParentFile();
		
		if (destDir == null || alwaysRename) {
			name = name.substring(0, name.length()-3);
		}	
		if (destDir == null) {
			return new File(parent, name);
		} else {
			return new File(destDir, name);
		}
	}
	
	protected File getOutfile(URL in) {
		String name = in.getFile();
		String newName = name.substring(0, name.length()-3);
		
//		msg("\nwritting '"+name+"' to '"+newName+"' in directory '"+
//			(destDir == null ? ".": destDir.toString())+"'");
		return new File(destDir, newName);
	}
	
	protected Reader sourceReader() {

		InputStream sis = new SequenceInputStream(new ISEnumeration());	
		
		return new InputStreamReader(sis);
	}
/*	
	protected Set sourceReaders() {

		Set readers = new HashSet();
	
		for (Iterator i = sources.iterator(); i.hasNext(); ) {
			try {
				readers.add(new InputStreamReader(createIS(i.next())));
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		}
		return readers;
	}
*/	
	protected class ISEnumeration implements Enumeration {
		Iterator srcIterator;
		public ISEnumeration() {
			srcIterator = sources.iterator();
		}
		public boolean hasMoreElements() {
			return srcIterator.hasNext();
		}
		public Object nextElement() {
			try {
				return createIS(srcIterator.next());
			} catch (IOException ioe) { return null;}
		}
	}
	
	protected static InputStream createIS(Object o) throws IOException {
		if (o instanceof File) 
			return new FileInputStream((File)o);
		else if (o instanceof URL)
			return ((URL)o).openStream();
		else
			return null;
	}
	
	protected void process() throws IOException {
		Writer writer = stdout ? new OutputStreamWriter(System.out) : null;
		Reader reader = null;
		if(concat) {
			msg("\n\nStart processing concatenated sources");
			reader = sourceReader();
			if (stdout || outFile == null) {
				msg(" -> stdout...");
			} else {
				msg(" -> file '"+outFile+"'...");
				writer = new FileWriter(outFile);
			}
			process(reader, writer);
			if (!stdout) 
				writer.close();
			reader.close();
			
		} else {
			msg("\n\nStart processing separately...");
			
			for (Iterator i = sources.iterator(); i.hasNext();) {
				Object o = i.next();
				if (o instanceof File) {
					File f = (File)o;
					reader = new FileReader(f);
					if (stdout) {
						msg("\n\tprocessing file '"+f+"' -> stdout");
					} else {
						msg("\n\tprocessing file '"+f+"' -> '"+getOutfile(f)+"'");
						writer = new FileWriter(getOutfile(f));
					}
				} else if (o instanceof URL) {
					URL u = (URL)o;
					reader = new InputStreamReader(u.openStream());
					if (stdout) {
						msg("\n\tprocessing URL '"+u+"' -> stdout");
					} else {
						msg("\n\tprocessing URL '"+u+"' -> '"+getOutfile(u)+"'");
						writer = new FileWriter(getOutfile(u));
					}
				} 
				process(reader, writer);
				if (!stdout) 
					writer.close();
				reader.close();
			}
		}
		if (stdout && writer != null) 
			writer.close();
		msg("\n...done\n");
	}
	
	public void process(Reader in, Writer o) throws IOException {
		
		out = o;

        CharReader r = new CharReader(readContent(in));

        EStack es = new EStack();
        Element e = null;
        boolean emptyElt = false;
        boolean isInText = true;
        int depth = 0;
        int textBegin = r.index();
        int afterName = 0;

       try {
        while(!r.eof()) {
            textBegin = r.index();
            char c = r.next();
            if (c == '%' && (isNamechar(r.at()) || r.at() == '%')) {
                //boolean goOut = false;
                do {
                    int afterPercent = r.index();
                    if (isInText && afterPercent-1 > textBegin) {
//                    	System.out.println("storing text["+textBegin+".."+(afterPercent-1)+"]="
//                    		+r.substring(textBegin, afterPercent-1));
                        text(r.substring(textBegin, afterPercent-1));
                    }
                    isInText = false;
                    c = r.next();
                    if (c == '%') {
                    	
                        // %% - empty element
//                        System.out.println("%% reached at "+(r.index()-1));
                        emptyElt = true;
                        c = r.next();
	                    while(!r.eof() && isNamechar(c)) {
	                        c = r.next();    
	                    }
	                    afterName = r.eof() ? r.index()+1 : r.index();
    	                e = new Element(r.substring(afterPercent+1, afterName-1), depth);
                        e.emptyTag();
                        
                        r.pushback();
                        
                        //System.out.println("after %% reached="+r.at());

                        closeNesting(es, e.depth());
                  //      goOut = true;
                        
                        //System.out.println("nesting closed");

	                    if(markup(r.at())) {
                            if(r.at() == '{')
                                isInText = true;
                        } else {
                            isInText = true;
                        }
                        
                        if(isInText) {
                            textBegin = r.index();
                        } else {
                            //r.pushback();
                            //System.out.println("not in text->step back="+r.at());
                        }
                        break;
                        
                    } else if (isNamechar(c)) {
                    	
                    	// nonempty element
                        c = r.next();
	                    while(!r.eof() && isNamechar(c)) {
	                        c = r.next();    
	                    }
    	                afterName = r.eof() ? r.index()+1 : r.index();
	                    e = new Element(r.substring(afterPercent, afterName-1), depth);
                        es.push(e);
                        depth++;

                        // to the next nonspace -> it will be attribute name
	                    c = r.nextNonspace(c);   
//	                    System.out.println("nextNonspace in the nonempty elt="+c);                 
                        if (c == '{' || c == '%') {
                            // attrs are finished or nested tag encountered
                            e.startTag();
                        } else {
                            while(!r.eof() && c != '{' && c != '%') {
                                int anBegin = r.index()-1;
                                while(!r.eof() && c != '=' && !isWhitespace(c)) {
                                    c = r.next();
                                }
                                /*
                                System.out.println("anBegin="+r.at(anBegin));
                                System.out.println("anEnd+1="+(r.index()-1));
                                System.out.println("an="+r.substring(anBegin, r.index()-1));
                                */
                                e.newAttrName(r.substring(anBegin, r.index()-1));
                                
                                
                                while(!r.eof() && !isQuote(c)) {
                                    c = r.next();
                                }
                                
                                r.pushback();
                                
                                char quote = r.next();
//                                System.out.println("quote="+quote);
                                
                                c = ' ';
                                int avBegin = r.index();
                                while(!r.eof() && c != quote) {
                                    c = r.next();
                                }
								/*
                                System.out.println("avBegin="+r.at(avBegin));
                                System.out.println("avEnd+1="+(r.index()-1));
                                System.out.println("av="+r.substring(avBegin, r.index()-1));
                                */
                                e.newAttrValue(r.substring(avBegin, r.index()-1));
                                
                                c = r.next();
        	                    c = r.nextNonspace(c);   
                                
                            }
    //                        if (!r.eof()) {
                                // attrs are finished or nested tag encountered
                                e.startTag();
    //                        }
                        }    

	                    // attrs to the elt on top.                
                        //e.startTag();
                    } else {
                    	// percent but not an element name
	                    isInText = true;
	    	            // c = r.next();
                	}
                } while(c == '%' && !r.eof()); // && !goOut
                
                depth = 0;
                
            } else if (c == '{') {
                e = new Brackets();
                es.push(e);
                e.startTag();
                
            } else if (c == '}') {
                closeNesting(es);
/*                
                // popping elts
                Element ee = null;
                if (!es.isEmpty()) {
	                do {
	                    ee = es.pop();
	                    ee.endTag();
                        System.out.println("...elt was popped, depth="+ee.depth()+", now size="+es.st.size());
	                } while (!es.isEmpty() && ee.depth() > 0);
				}				
				if (!es.isEmpty()) {
//					System.out.println("es.size="+es.st.size());
                    System.out.println("popping another...");
					es.pop().endTag();
				}	
*/				
                isInText = true;
                
            } else {
            	// normal char (not  % { }) or % and not followed by element name
                // System.out.println("normal char="+c+" at="+(r.index()-1)+", tb="+textBegin); 
				
				if (c == '%' && !r.eof()) {
					c = r.next();
				}
	
                while(!r.eof() && notMarkup(c)) {
                	c = r.next();
            	}
            	r.pushback();
            	int firstMarkup = r.eof() ? r.index()+1 : r.index();
            	text(r.substring(textBegin, firstMarkup));
//            	System.out.println("text="+r.substring(textBegin, firstMarkup-1));
            }
        }
        // popping elts
		while(!es.isEmpty()) {
			es.pop().endTag();
		}
        
     } catch(StringIndexOutOfBoundsException aioob) {
		while(!es.isEmpty()) {
			es.pop().endTag();
		}
	 }	
//	 out.close();
    }

    public static void closeNesting(EStack es, int depth) throws IOException {
        while (!es.isEmpty() && depth > 0) {
            es.pop().endTag();
            depth--;
        }
    }
    
    public static void closeNesting(EStack es) throws IOException {
        while (!es.isEmpty() && es.top().depth() > 0) {
            es.pop().endTag();
        }
        if (!es.isEmpty()) es.pop().endTag();
    }
    
    public static boolean markup(char c) {
        return (c == '%' ) || c == '{' || c == '}'; // && (isNamechar(next) || next == '%')
    }

    public static boolean notMarkup(char c) {
        return !markup(c);
    }

    public static boolean isNamechar(char c) {
        return Character.isLetterOrDigit(c) || c == '_' || c == ':' || c == '-' || c == '.';
    }

    public static boolean isWhitespace(char c) {
        return Character.isWhitespace(c);
    }

    public static boolean isQuote(char c) {
        return c == '\'' || c == '"';
    }

    public static boolean isTagMark(char c) {
        return c == '%';
    }

    public void msg(String s) {
    	if (verbose)
        	System.err.print(s);
    }
    
    public void text(String s) throws IOException {
        out.write(s);
    }
    
    public static String readContent(Reader in) {
        try {
            StringWriter wr = new StringWriter();

            int oneChar;
            while ((oneChar=in.read()) != -1)
                wr.write(oneChar);
            in.close();
            wr.close();
            return wr.toString();
        } catch (IOException ioe) {
            return "";
        }
    }

protected class Element {

    private String name;
    private Map attrs;
    
    private String attrName, attrValue;
    private char avDelim;
    private int depth;

    public Element(String name) {
        attrs = new HashMap();
        this.name = name;
    }
    
    public Element(String name, int d) {
        this(name);
        this.depth = d;
//        System.out.println("creating elt="+name+", depth="+d);
    }
    
    public void newAttrName(String name) {
        attrName = name;
    }
    
    public void newAttrValue(String value) {
        attrValue = value;
        addAttribute();
    }
    
    protected void addAttribute() {
        attrs.put(attrName, attrValue);
//        System.out.println("attrs size="+attrs.size());
    }
    
    public void startTag() throws IOException {
        out.write("<"+name+(attrs.keySet().isEmpty() ? "" : " "));
        for (Iterator i = attrs.keySet().iterator(); i.hasNext(); ) {
            String k = (String)i.next();
            String av = (String)attrs.get(k);
            if (av.indexOf('"') >= 0)
            	out.write(k+"='"+av+"' ");
            else 
            	out.write(k+"=\""+av+"\" ");
        }
        out.write(">");
    }
    
    public void endTag() throws IOException {
        out.write("</"+name+">");
    }
    
    public void emptyTag() throws IOException {
        out.write("<"+name+"/>");
    }
    
    public int depth() {
        return depth;
    }
    
    public String getName() {
        return name;
    }
}    

protected class Brackets extends Element {
    public Brackets() {
        super("{");
    }
    public void startTag() throws IOException {
        out.write('{');
    }
    
    public void endTag() throws IOException {
        out.write('}');
    }
    
    public void emptyTag() throws IOException {
        out.write('{');
    }
    
}

protected class EStack {
    List st = new ArrayList();
    
    public void push(Element e) {
//    	System.err.println("pushing "+e.getName()+" depth="+e.depth());
        st.add(e);
    }
    
    public Element pop() {
        int top = st.size()-1;
        Element e = (Element)st.remove(top);
        return e;
    }    
    
    public Element top() {
        int top = st.size()-1;
        Element e = (Element)st.get(top);
        return e;
    }    
    
    public boolean isEmpty() {
        return st.isEmpty();
    }
}

// end of class XPP
}

class CharReader {
    private String src; // = "%html{%head{}%body{ %tag1 attr1='attr1value' attr2=\"\"{} %tag2 attr3=\"attr3value\"{Obsah tag2}  %tag4%enclosedtag4 attr4=\"attr4value\"{}}}";

    private int index;
    private int len;
    private boolean eof = false;
    
    public CharReader(String s) { src = s; len = s.length(); }

    public char next() {
    	if (index < len) {
    		eof = false;
        	return src.charAt(index++);
    	} else {
    		eof = true;
    		return 0;
    	}
    }

    public char at() {
        return at(index);
    }

    public char at(int i) {
    	if (i < len) {
        	return src.charAt(i);
    	} else {
    		return 0;
    	}
    }

	public boolean eof() { return eof; }
	
    public char nextNonspace(char c) {
        while(!eof() && Character.isWhitespace(c)) {
            c = next();
        }
        return c;
    }

    public int index() {
        return index;
    }

    public char pushback() {
        index--;
        return at();
    }

    public String substring(int textBegin, int textEnd) {
        return src.substring(textBegin, textEnd);
    }
}

